/*
 * livepatch_bsc1200058
 *
 * Fix for CVE-2022-1882, bsc#1200058
 *
 *  Upstream commit:
 *  353f7988dd84 ("watchqueue: make sure to serialize 'wqueue->defunct
 *                 properly")
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  Not affected
 *
 *  SLE15-SP2 and -SP3 commit:
 *  Not affected
 *
 *  SLE15-SP4 commit:
 *  0e9524c9c5aec914a055ec3dbda8012da22a3de6
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Marcos Paulo de Souza <mpdesouza@suse.com>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from kernel/watch_queue.c */
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/printk.h>

/* klp-ccp: from kernel/watch_queue.c */
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

/* klp-ccp: from include/linux/preempt.h */
#ifdef CONFIG_PREEMPT_COUNT
#ifdef CONFIG_PREEMPTION

#ifdef CONFIG_X86_64
/* klp-ccp: from arch/x86/include/asm/preempt.h */
#ifdef CONFIG_PREEMPT_DYNAMIC
static void (*klpe___SCT__preempt_schedule)(void);

#define klpr___preempt_schedule() \
	(*klpe___SCT__preempt_schedule)()

#else /* CONFIG_PREEMPT_DYNAMIC */
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#else /* CONFIG_X86_64 */
#error "klp-ccp: a preceeding branch should have been taken"
#endif

/* klp-ccp: from include/linux/preempt.h */
#define klpr_preempt_enable() \
do { \
	barrier(); \
	if (unlikely(preempt_count_dec_and_test())) \
		klpr___preempt_schedule(); \
} while (0)

#else /* CONFIG_PREEMPTION */
#error "klp-ccp: a preceeding branch should have been taken"
#endif

#else /* CONFIG_PREEMPT_COUNT */
#define klpr_preempt_enable()			barrier()
#endif

/* klp-ccp: from include/linux/highmem-internal.h */
#ifdef CONFIG_HIGHMEM
#error "klp-ccp: non-taken branch"
#else /* CONFIG_HIGHMEM */

static inline void klpr___kunmap_atomic(void *addr)
{
#ifdef ARCH_HAS_FLUSH_ON_KUNMAP
#error "klp-ccp: non-taken branch"
#endif
	pagefault_enable();
	if (IS_ENABLED(CONFIG_PREEMPT_RT))
		migrate_enable();
	else
		klpr_preempt_enable();
}
#endif /* CONFIG_HIGHMEM */



/* klp-ccp: from kernel/watch_queue.c */
#include <linux/eventpoll.h>
#include <linux/uaccess.h>

/* klp-ccp: from include/linux/security.h */
struct watch_notification;

static int (*klpe_security_post_notification)(const struct cred *w_cred,
			       const struct cred *cred,
			       struct watch_notification *n);

/* klp-ccp: from kernel/watch_queue.c */
#include <linux/cred.h>
#include <linux/sched/signal.h>
#include <linux/watch_queue.h>

/* klp-ccp: from kernel/watch_queue.c */
#include <linux/pipe_fs_i.h>

#define WATCH_QUEUE_NOTE_SIZE 128
#define WATCH_QUEUE_NOTES_PER_PAGE (PAGE_SIZE / WATCH_QUEUE_NOTE_SIZE)

static const struct pipe_buf_operations (*klpe_watch_queue_pipe_buf_ops);

/*
 * This must be called under the RCU read-lock, which makes
 * sure that the wqueue still exists. It can then take the lock,
 * and check that the wqueue hasn't been destroyed, which in
 * turn makes sure that the notification pipe still exists.
 */
static inline bool lock_wqueue(struct watch_queue *wqueue)
{
	spin_lock_bh(&wqueue->lock);
	if (unlikely(wqueue->defunct)) {
		spin_unlock_bh(&wqueue->lock);
		return false;
	}
	return true;
}

static inline void unlock_wqueue(struct watch_queue *wqueue)
{
	spin_unlock_bh(&wqueue->lock);
}

static bool klpp_post_one_notification(struct watch_queue *wqueue,
				  struct watch_notification *n)
{
	void *p;
	struct pipe_inode_info *pipe = wqueue->pipe;
	struct pipe_buffer *buf;
	struct page *page;
	unsigned int head, tail, mask, note, offset, len;
	bool done = false;

	if (!pipe)
		return false;

	spin_lock_irq(&pipe->rd_wait.lock);

	mask = pipe->ring_size - 1;
	head = pipe->head;
	tail = pipe->tail;
	if (pipe_full(head, tail, pipe->ring_size))
		goto lost;

	note = find_first_bit(wqueue->notes_bitmap, wqueue->nr_notes);
	if (note >= wqueue->nr_notes)
		goto lost;

	page = wqueue->notes[note / WATCH_QUEUE_NOTES_PER_PAGE];
	offset = note % WATCH_QUEUE_NOTES_PER_PAGE * WATCH_QUEUE_NOTE_SIZE;
	get_page(page);
	len = n->info & WATCH_INFO_LENGTH;
	p = kmap_atomic(page);
	memcpy(p + offset, n, len);
	klpr___kunmap_atomic(p);

	buf = &pipe->bufs[head & mask];
	buf->page = page;
	buf->private = (unsigned long)wqueue;
	buf->ops = &(*klpe_watch_queue_pipe_buf_ops);
	buf->offset = offset;
	buf->len = len;
	buf->flags = PIPE_BUF_FLAG_WHOLE;
	smp_store_release(&pipe->head, head + 1); /* vs pipe_read() */

	if (!test_and_clear_bit(note, wqueue->notes_bitmap)) {
		spin_unlock_irq(&pipe->rd_wait.lock);
		BUG();
	}
	wake_up_interruptible_sync_poll_locked(&pipe->rd_wait, EPOLLIN | EPOLLRDNORM);
	done = true;

out:
	spin_unlock_irq(&pipe->rd_wait.lock);
	if (done)
		kill_fasync(&pipe->fasync_readers, SIGIO, POLL_IN);
	return done;

lost:
	buf = &pipe->bufs[(head - 1) & mask];
	buf->flags |= PIPE_BUF_FLAG_LOSS;
	goto out;
}

static bool filter_watch_notification(const struct watch_filter *wf,
				      const struct watch_notification *n)
{
	const struct watch_type_filter *wt;
	unsigned int st_bits = sizeof(wt->subtype_filter[0]) * 8;
	unsigned int st_index = n->subtype / st_bits;
	unsigned int st_bit = 1U << (n->subtype % st_bits);
	int i;

	if (!test_bit(n->type, wf->type_filter))
		return false;

	for (i = 0; i < wf->nr_filters; i++) {
		wt = &wf->filters[i];
		if (n->type == wt->type &&
		    (wt->subtype_filter[st_index] & st_bit) &&
		    (n->info & wt->info_mask) == wt->info_filter)
			return true;
	}

	return false; /* If there is a filter, the default is to reject. */
}

void klpp___post_watch_notification(struct watch_list *wlist,
			       struct watch_notification *n,
			       const struct cred *cred,
			       u64 id)
{
	const struct watch_filter *wf;
	struct watch_queue *wqueue;
	struct watch *watch;

	if (((n->info & WATCH_INFO_LENGTH) >> WATCH_INFO_LENGTH__SHIFT) == 0) {
		WARN_ON(1);
		return;
	}

	rcu_read_lock();

	hlist_for_each_entry_rcu(watch, &wlist->watchers, list_node) {
		if (watch->id != id)
			continue;
		n->info &= ~WATCH_INFO_ID;
		n->info |= watch->info_id;

		wqueue = rcu_dereference(watch->queue);
		wf = rcu_dereference(wqueue->filter);
		if (wf && !filter_watch_notification(wf, n))
			continue;

		if ((*klpe_security_post_notification)(watch->cred, cred, n) < 0)
			continue;

		if (lock_wqueue(wqueue)) {
			klpp_post_one_notification(wqueue, n);
			unlock_wqueue(wqueue);
		}
	}

	rcu_read_unlock();
}

static void (*klpe_free_watch)(struct rcu_head *rcu);

static void klpr___put_watch(struct kref *kref)
{
	struct watch *watch = container_of(kref, struct watch, usage);

	call_rcu(&watch->rcu, (*klpe_free_watch));
}

static void klpr_put_watch(struct watch *watch)
{
	kref_put(&watch->usage, klpr___put_watch);
}

int klpp_add_watch_to_object(struct watch *watch, struct watch_list *wlist)
{
	struct watch_queue *wqueue = rcu_access_pointer(watch->queue);
	struct watch *w;

	hlist_for_each_entry(w, &wlist->watchers, list_node) {
		struct watch_queue *wq = rcu_access_pointer(w->queue);
		if (wqueue == wq && watch->id == w->id)
			return -EBUSY;
	}

	watch->cred = get_current_cred();
	rcu_assign_pointer(watch->watch_list, wlist);

	if (atomic_inc_return(&watch->cred->user->nr_watches) >
	    task_rlimit(current, RLIMIT_NOFILE)) {
		atomic_dec(&watch->cred->user->nr_watches);
		put_cred(watch->cred);
		return -EAGAIN;
	}

	if (lock_wqueue(wqueue)) {
		kref_get(&wqueue->usage);
		kref_get(&watch->usage);
		hlist_add_head(&watch->queue_node, &wqueue->watches);
		unlock_wqueue(wqueue);
	}

	hlist_add_head(&watch->list_node, &wlist->watchers);
	return 0;
}

int klpp_remove_watch_from_object(struct watch_list *wlist, struct watch_queue *wq,
			     u64 id, bool all)
{
	struct watch_notification_removal n;
	struct watch_queue *wqueue;
	struct watch *watch;
	int ret = -EBADSLT;

	rcu_read_lock();

again:
	spin_lock(&wlist->lock);
	hlist_for_each_entry(watch, &wlist->watchers, list_node) {
		if (all ||
		    (watch->id == id && rcu_access_pointer(watch->queue) == wq))
			goto found;
	}
	spin_unlock(&wlist->lock);
	goto out;

found:
	ret = 0;
	hlist_del_init_rcu(&watch->list_node);
	rcu_assign_pointer(watch->watch_list, NULL);
	spin_unlock(&wlist->lock);

	/* We now own the reference on watch that used to belong to wlist. */

	n.watch.type = WATCH_TYPE_META;
	n.watch.subtype = WATCH_META_REMOVAL_NOTIFICATION;
	n.watch.info = watch->info_id | watch_sizeof(n.watch);
	n.id = id;
	if (id != 0)
		n.watch.info = watch->info_id | watch_sizeof(n);

	wqueue = rcu_dereference(watch->queue);

	if (lock_wqueue(wqueue)) {
		klpp_post_one_notification(wqueue, &n.watch);

		if (!hlist_unhashed(&watch->queue_node)) {
			hlist_del_init_rcu(&watch->queue_node);
			klpr_put_watch(watch);
		}

		unlock_wqueue(wqueue);
	}

	if (wlist->release_watch) {
		void (*release_watch)(struct watch *);

		release_watch = wlist->release_watch;
		rcu_read_unlock();
		(*release_watch)(watch);
		rcu_read_lock();
	}
	klpr_put_watch(watch);

	if (all && !hlist_empty(&wlist->watchers))
		goto again;
out:
	rcu_read_unlock();
	return ret;
}



#include <linux/kernel.h>
#include "livepatch_bsc1200058.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
#ifdef CONFIG_PREEMPTION
	{ "__SCT__preempt_schedule", (void *)&klpe___SCT__preempt_schedule },
#endif
	{ "free_watch", (void *)&klpe_free_watch },
	{ "security_post_notification",
	  (void *)&klpe_security_post_notification },
	{ "watch_queue_pipe_buf_ops", (void *)&klpe_watch_queue_pipe_buf_ops },
};

int livepatch_bsc1200058_init(void)
{
	return klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}

